// mod -> header file for GBA

#include <io.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//---------------------------------------------------------------------------

#define TURE  1
#define FALSE 0

#define MAX_SAMPLE_DATAS 16384

#define INVALID_SMP_DATA 0xffffffff

#define MOD_NO_NOTE      63
#define MOD_NO_SAMPLE    31


#define ASSERT(cond)                                                                 \
	if((cond) == FALSE) {                                                            \
		fprintf(stdout, "\nASSERT FAILED\n\t%s\n\tLine %i\n\n", __FILE__, __LINE__); \
		exit(1);                                                                     \
	}                                                                                \

//---------------------------------------------------------------------------

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned long u32;

typedef signed char s8;
typedef signed short s16;
typedef signed long s32;

//---------------------------------------------------------------------------

//TvOfW^f[^
typedef struct
{
	s8* data;
	u32 length;
} ST_SAMPLE_DATA;

//F
typedef struct
{
	u8  name[22];
	u16 length;
	u8  finetune;
	u8  vol;
	u16 loopStart;
	u16 loopLength;
	u32 smpDataIdx;
} ST_SAMPLE_HEADER;

//MODt@Cwb_
typedef struct
{
	u8               name[20];
	ST_SAMPLE_HEADER sample[31];
	u8               order[128];
	u8**             pattern;
	u8               orderCount;
	u8               patternCount;
	u32              smpDataCount;
	ST_SAMPLE_DATA   smpData[MAX_SAMPLE_DATAS];
} ST_MOD_HEADER;

//---------------------------------------------------------------------------
const u16 periodTable[12*5] =
{
	1712, 1616, 1524, 1440, 1356, 1280, 1208, 1140, 1076, 1016,  960, 906,
	 856,  808,  762,  720,  678,  640,  604,  570,  538,  508,  480, 453,
	 428,  404,  381,  360,  339,  320,  302,  285,  269,  254,  240, 226,
	 214,  202,  190,  180,  169,  160,  151,  142,  134,  127,  120, 113,
	 107,  101,   95,   90,   84,   80,   75,   71,   67,   63,   60,  56,
};
//---------------------------------------------------------------------------
ST_MOD_HEADER MOD;

//---------------------------------------------------------------------------
void LoadMod        (char* fileName);
void LoadSamples    (FILE* modFile);
void LoadOrders     (FILE* modFile);
void LoadPatterns   (FILE* modFile);
void LoadSampleDatas(FILE* modFile);

void WriteHeaderFile(char* fileName);
//---------------------------------------------------------------------------
int main(int argc, char **argv)
{
	if( argc != 2 )
	{
		fprintf(stdout, "Usage: mod2h [filename]\n");
		exit(1);
	}

	memset(&MOD, 0, sizeof(ST_MOD_HEADER));

	LoadMod(argv[1]);
	WriteHeaderFile("ModData.h");

	fprintf(stdout, "Success!\n");
	return 0;
}

//---------------------------------------------------------------------------

void LoadMod(char* fileName)
{
	FILE* modFile;

	modFile = fopen(fileName, "rb");
	if(modFile == NULL)
	{
		fprintf(stderr, "%s does not exist\n", fileName);
		exit(1);
	}

	//MODt@C`FbNAgbN`FbN
	u8 sig[4];
	fseek(modFile, 1080, SEEK_SET);
	fread(sig, 4, 1, modFile);
	if( memcmp(sig, "M.K.", 4) && memcmp(sig, "4CHN", 4) )
	{
		fprintf(stderr, "%s is not a 4-channel MOD file\n", fileName);
		exit(1);
	}

	//MODt@Ĉœǂݍ݂߂Jn
	fseek(modFile, 0, SEEK_SET);

	//^Cgǂݍ
	fread(MOD.name, 20, 1, modFile);
	MOD.name[19] = '\0';

	LoadSamples     (modFile);
	LoadOrders      (modFile);
	LoadPatterns    (modFile);
	LoadSampleDatas (modFile);

	fclose(modFile);
}

//---------------------------------------------------------------------------
void LoadSamples(FILE* modFile)
{
	u32 i;
	for(i=0; i<31; i++)
	{
		ST_SAMPLE_HEADER* smp = &MOD.sample[i];
		fread(smp, 30, 1, modFile);

		smp->smpDataIdx = INVALID_SMP_DATA;
		smp->length     = ( ((smp->length     & 0xff) << 8) | (smp->length     >> 8) );
		smp->loopStart  = ( ((smp->loopStart  & 0xff) << 8) | (smp->loopStart  >> 8) );
		smp->loopLength = ( ((smp->loopLength & 0xff) << 8) | (smp->loopLength >> 8) );
	}
}
//---------------------------------------------------------------------------
void LoadOrders(FILE* modFile)
{
	u32 trash;
	fread(&MOD.orderCount,   1, 1, modFile);
	fread(&trash         ,   1, 1, modFile);	// Unused byte
	fread(&MOD.order     , 128, 1, modFile);
	fread(&trash         ,   4, 1, modFile);	// Unused byte
}
//---------------------------------------------------------------------------
void LoadPatterns(FILE* modFile)
{
	u32 i;
	u8 highestOrder = 0;


	//p^[̐߂
	for(i=0; i<MOD.orderCount; i++)
	{
		if(MOD.order[i] > highestOrder)
		{
			highestOrder = MOD.order[i];
		}
	}
	//0JnȂ̂+1Ƃ
	MOD.patternCount = highestOrder + 1;


	//p^[̐|C^m
	MOD.pattern = new u8*[MOD.patternCount];
	ASSERT(MOD.pattern);
	memset(MOD.pattern, 0, MOD.patternCount * sizeof(u8*));


	u8 curPattern, row, column;
	for(curPattern=0; curPattern<MOD.patternCount; curPattern++)
	{
		//p^[̈̊m
		MOD.pattern[curPattern] = new u8[1024];
		ASSERT(MOD.pattern[curPattern]);
		memset(MOD.pattern[curPattern], 0, 1024);

		//4*64̃p^[ǂݍ
		for(row=0; row<64; row++)
		{
			for(column=0; column<4; column++)
			{
				u8  cell[4];
				u8  sample;
				u16 period;
				u8  effect;
				u8  param;
				u8  closestNote;

				fread(cell, 4, 1, modFile);

				sample = (cell[0] & 0xf0) | (cell[2] >> 4);
				period = (cell[1]) | ((cell[0] & 0x0f) << 8);
				effect = cell[2] & 0x0f;
				param  = cell[3];

				if(sample == 0) sample = MOD_NO_SAMPLE;
				else            sample -= 1;

				if(period == 0)
				{	// Period 0 is no note
					closestNote = MOD_NO_NOTE;
				}
				else
				{
					u16 closestDist = 0xffff;
					closestNote = 0;

					// 5 octaves, 12 notes each
					for(i=0; i<12*5; i++)
					{
						u16 newDist = abs( period - periodTable[i] );
						if(newDist < closestDist)
						{
							closestNote = (u8)i;
							closestDist = newDist;
						}
					}
				}

				u8 *outCell = &MOD.pattern[curPattern][row*4*4 + column*4];
				outCell[0]  = closestNote;
				outCell[1]  = sample;
				outCell[2]  = effect;
				outCell[3]  = param;
			} // for cloum
		} // for row
	} // for curPattern
}
//---------------------------------------------------------------------------
void LoadSampleDatas(FILE *modFile)
{
	MOD.smpDataCount = 0;
	
	u32 i;
	u16 length;
	for(i=0; i<31; i++)
	{
		length = MOD.sample[i].length * 2;
		
		if(length == 0)
		{
			continue;
		}

		MOD.smpData[MOD.smpDataCount].length = length;
		MOD.smpData[MOD.smpDataCount].data   = new s8[length];
		ASSERT(MOD.smpData[MOD.smpDataCount].data != NULL);

		fread (MOD.smpData[MOD.smpDataCount].data, length, 1, modFile);

		MOD.sample[i].smpDataIdx = MOD.smpDataCount;
		MOD.smpDataCount++;
	}
}

//---------------------------------------------------------------------------
void WriteHeaderFile(char* fileName)
{
	u32 i,j;
	FILE* outFile;

	outFile = fopen(fileName, "wt");
	if(outFile == NULL)
	{
		fprintf(stderr, "Failed to create %s\n", fileName);
		exit(1);
	}


	// header
	fprintf(outFile, "// MOD -> HeaderFile Converter\n\n");
	fprintf(outFile, "#include \"header\\gba.h\"\n");
	fprintf(outFile, "#include \"Mod2H.h\"\n\n");

	// pattern
	for(i=0; i<MOD.patternCount; i++)
	{
		fprintf(outFile, "const u8 dModPattern%i[] = {", i);

		for(j=0; j<1024; j++)
		{
			if((j & 15) == 0)	//sp
			{
				fprintf(outFile, "\n\t");
			}
			fprintf(outFile, "%3i, ", MOD.pattern[i][j]);
		}
		fprintf(outFile, "\n};\n");
	}
	fprintf(outFile, "\n");


	// SampleDatas
	for(i=0; i<MOD.smpDataCount; i++)
	{
		fprintf(outFile, "const s8 dSmpData%i[] = {", i);

		for(j=0; j<MOD.smpData[i].length; j++)
		{
			if((j & 15) == 0)	//sp
			{
				fprintf(outFile, "\n\t");
			}
			fprintf(outFile, "%3i, ", MOD.smpData[i].data[j]);
		}
		fprintf(outFile, "\n};\n");
	}
	fprintf(outFile, "\n");


	// SampleTable
	fprintf(outFile, "const ST_SAMPLE_HEADER dModSmpTable[] = {");
	for(i=0; i<MOD.smpDataCount; i++)
	{
		fprintf( outFile, "\n\t{ %i, %i, %i, %i, %i, dSmpData%i }, ", 
				MOD.sample[i].length, 
				MOD.sample[i].finetune, 
				MOD.sample[i].vol, 
				MOD.sample[i].loopStart, 
				MOD.sample[i].loopLength, 
				MOD.sample[i].smpDataIdx );
	}
	fprintf(outFile, "\n};\n");


	// PatternTable
	fprintf(outFile, "\nconst u8* dModPatternTable[] = {");
	for(i=0; i<MOD.patternCount; i++)
	{
		if((i & 4) == 0)
		{
			fprintf(outFile, "\n\t");
		}
		fprintf(outFile, "dModPattern%i, ", i);
	}
	fprintf(outFile, "\n};\n");


	// PatternTable
	fprintf(outFile, "\nconst u8 dModOrderTable[] = {");
	for(i=0; i<MOD.orderCount; i++)
	{
		if((i & 15) == 0)
		{
			fprintf(outFile, "\n\t");
		}
		fprintf(outFile, "%i, ", MOD.order[i]);
	}
	fprintf(outFile, "\n};\n");

	// dModTable
	fprintf(outFile, "\nconst ST_MOD_HEADER dModTable = {");
	fprintf(outFile, "dModSmpTable, ");
	fprintf(outFile, "dModOrderTable, ");
	fprintf(outFile, "dModPatternTable, ");
	fprintf(outFile, "%i, {",   MOD.orderCount);
	fprintf(outFile, "%i,0,0} };", MOD.patternCount);
	fprintf(outFile, "\n");


	fclose(outFile);
}
